// killLimits.cpp : Defines the entry point for the DLL application.
//

#include <string>
#include <vector>
#include <map>
#include <fstream>
#include "bzfsAPI.h"

BZ_GET_PLUGIN_VERSION


class killLimits : public bz_EventHandler
{
public:

	struct limit
	{
		int lim;
		bool isrelative;
		std::string flag;
	};

	struct flagtracker
	{
		std::string flag;
		int kills;
		int killlimit; //used for relative kill limits (e.g. 2n)
	};

	virtual void process (bz_EventData *eventData);
	void readArg (std::string arg);

	std::vector<std::string>* explode (std::string src, char s);
	int strtoint (std::string src);
	void trim (std::string& str); //trims whitespace from the ends of a string
	void removeempties(std::vector<std::string>& arr); //removes empty string items in vector
	int totaltanks(); //count tanks
	std::string abbrev(std::string src); //convert player data flag string to abbreviation
	std::string lookupflag(int playerID); //look up the given player's flag

	std::vector<limit> limits; 
	std::map<int, flagtracker> playerflags; //one object per player

	bool usetotalgenokills;
	bool onlycountopponents; //only count opponents

	int debuglevel;

	bool bork; //if plugin encounters an error
};

killLimits kl;

BZF_PLUGIN_CALL int bz_Load ( const char* commandLine )
{
	kl.bork = false;
	kl.usetotalgenokills = false;
	kl.onlycountopponents = false;
	kl.debuglevel = 4;

	//open conf file
	std::string conffilename = commandLine;

	std::ifstream conf;
	conf.open(conffilename.c_str());

	if (!conf.is_open())
	{
		std::string errormsg = "killLimits: ERROR: failed to open file \"" + conffilename + "\"";
		bz_debugMessage(kl.debuglevel,errormsg.c_str());
		kl.bork = true;
	}
	else
	{
		//read the conf file
		std::string curline;
		std::vector<std::string>* curlineargs;

		while (getline(conf, curline))
		{
			//snip the comment out of the line if it exists
			std::string::size_type index = curline.find('#');
			if (index != std::string::npos)
			{
				curline = curline.substr(0, index);
			}
			//split the line into arguments
			curlineargs = kl.explode(curline, '-');

			//there should be nothing other than whitespace before the first hyphen... check.
			kl.trim(curlineargs->at(0));
			if (curlineargs->at(0) != "")
			{
				std::string errormsg = "killLimits: warning: invalid argument \"" + curlineargs->at(0) + "\" (ignoring)";
				bz_debugMessage(kl.debuglevel,errormsg.c_str());
			}

			for (unsigned int x = 1; x < curlineargs->size(); x++)
			{
				kl.trim(curlineargs->at(x));
				if (curlineargs->at(x) != "") //only pass the argument if it's something besides whitespace
				{
					kl.readArg(curlineargs->at(x));
				}
			}
			delete curlineargs;
		}
		conf.close();

		if (kl.limits.size() == 0)
		{
			std::string errormsg = "killLimits: ERROR: configuration file \"" + conffilename + "\" contains no kill limits";
			bz_debugMessage(kl.debuglevel,errormsg.c_str());
			kl.bork = true;
		}
	}

	if (!kl.bork)
	{
		bz_registerEvent (bz_ePlayerDieEvent, &kl);
		bz_registerEvent (bz_ePlayerJoinEvent, &kl);

		bz_debugMessage(kl.debuglevel,bz_format("killLimits plugin loaded: %d kill limits",kl.limits.size()));
	}
	else
	{
        bz_debugMessage(kl.debuglevel,"killLimits plugin was NOT loaded because of error(s)");
	}
	return 0;
}

BZF_PLUGIN_CALL int bz_Unload ( void )
{
	if (!kl.bork)	
	{
		bz_removeEvent (bz_ePlayerDieEvent, &kl);
		bz_removeEvent (bz_ePlayerJoinEvent, &kl);
		bz_debugMessage(kl.debuglevel,"killLimits plugin unloaded");
	}
	return 0;
}

void killLimits::process(bz_EventData* eventData)
{
	if (eventData->eventType == bz_ePlayerDieEvent)
	{
		bz_PlayerDieEventData* dat = (bz_PlayerDieEventData*)eventData;

		//reset the victim's stats
		playerflags[dat->playerID].flag = "";
		playerflags[dat->playerID].kills = 0;
		playerflags[dat->playerID].killlimit = 0;

		//deal with killer info

		if (dat->killerID == BZ_SERVER)
		{
			return;
		}

		bool directgeno = dat->flagKilledWith == "G";

		std::string flag = dat->flagKilledWith.c_str();
		if (flag == "") //sometimes it's misreported, so let's check and make sure
		{
			flag = lookupflag(dat->killerID);
		}
		else if (flag != lookupflag(dat->killerID))
		{
			return; //if the player doesn't actually have the flag, we don't need to do anything
		}

		if (flag == playerflags[dat->killerID].flag)
		{
			if (playerflags[dat->killerID].killlimit != 0)
			{
				if (flag != "G" || directgeno || usetotalgenokills)
				{
					playerflags[dat->killerID].kills++;
				}

				if (playerflags[dat->killerID].kills >= playerflags[dat->killerID].killlimit)
				{
					//met limit; zap flag and reset stats
					bz_removePlayerFlag(dat->killerID);
					playerflags[dat->killerID].flag = "";
					playerflags[dat->killerID].kills = 0;
					playerflags[dat->killerID].killlimit = 0;
				}
				else if (flag != "G" || directgeno || usetotalgenokills)
				{
					int killsleft = playerflags[dat->killerID].killlimit - playerflags[dat->killerID].kills;
					if (killsleft <= 5 || killsleft % 5 == 0)
					{
						if (killsleft == 1)
						{
							bz_sendTextMessage (BZ_SERVER, dat->killerID, "1 kill left");
						}
						else
						{
							bz_sendTextMessage (BZ_SERVER, dat->killerID, bz_format("%d kills left", killsleft)); 
						}
					}
				}
			}
		}
		else
		{
			//new flag, so we need to determine a kill limit (if it exists)
			playerflags[dat->killerID].flag = flag;
			playerflags[dat->killerID].kills = 1;
			playerflags[dat->killerID].killlimit = 0;

			//search for a kill limit...
			for (unsigned int x = 0; x < limits.size(); x++)
			{
				if (limits.at(x).flag == flag)
				{
					playerflags[dat->killerID].killlimit = limits.at(x).lim;
					if (limits.at(x).isrelative)
					{
						int mult = totaltanks();

						if (onlycountopponents)
						{
							//look up team and subtract teammates if they're on a colored team
							bz_PlayerRecord* playerinfo = bz_getPlayerByIndex(dat->killerID);
							if (playerinfo != NULL)
							{
								bz_eTeamType team = playerinfo->team;
								if (team == eRedTeam || team == eGreenTeam || team == eBlueTeam || team == ePurpleTeam)
								{
									mult -= bz_getTeamCount(team);
								}


								bz_freePlayerRecord (playerinfo);
							}
						}

						playerflags[dat->killerID].killlimit *= mult;
					}
					if (playerflags[dat->killerID].killlimit > 2)
					{
						bz_sendTextMessage (BZ_SERVER, dat->killerID, bz_format("%d kills left", playerflags[dat->killerID].killlimit - 1)); 
					}
					else if (playerflags[dat->killerID].killlimit == 2)
					{
						bz_sendTextMessage (BZ_SERVER, dat->killerID, "1 kill left");
					}
					break;
				}
			}
		}
	}
	else if (eventData->eventType == bz_ePlayerJoinEvent)
	{
		//clear stats for a new player (housekeeping)
		bz_PlayerJoinPartEventData* dat = (bz_PlayerJoinPartEventData*)eventData;
		playerflags[dat->playerID].flag = "";
		playerflags[dat->playerID].kills = 0;
		playerflags[dat->playerID].killlimit = 0;
	}
}

void killLimits::readArg(std::string arg)
{
	std::vector<std::string>* params = explode(arg, ' '); //split the argument into params
	removeempties(*params); //remove empty string elements (in case of multiple spaces)

	if (params->size() == 0)
	{
		return; //stop if the vector is empty... this should not happen
	}

	//read the arguments, now

	if (params->at(0) == "usetotalgenokills")
	{
		usetotalgenokills = true;
	}
	else if (params->at(0) == "debuglevel")
	{
		if (params->size() != 2)
		{
			std::string errormsg = "killLimits: warning: malformed parameter \"-" + arg + "\" (ignoring)";
			bz_debugMessage(debuglevel,errormsg.c_str());
		}
		else
		{
			int level = strtoint(params->at(1));
			if (level <= 4 && level >= 1)
			{
				debuglevel = level;
			}
		}
	}
	else if (params->at(0) == "onlycountopponents")
	{
		onlycountopponents = true;
	}
	else if (params->at(0) == "kl")
	{
		if (params->size() != 3)
		{
			std::string errormsg = "killLimits: warning: malformed parameter \"-" + arg + "\" (ignoring)";
			bz_debugMessage(debuglevel,errormsg.c_str());
		}
		else
		{
			limit newlimit;
			newlimit.flag = params->at(1);

			if (params->at(2)[params->at(2).length() - 1] == 'n')
			{
				newlimit.isrelative = true;
				params->at(2) = params->at(2).substr(0, params->at(2).length() - 1);
			}
			else
			{
				newlimit.isrelative = false;
			}

			int lim = strtoint(params->at(2));

			newlimit.lim = lim;

			if (lim == 0)
			{
				std::string errormsg = "killLimits: warning: malformed parameter \"-" + arg + "\" (ignoring)";
				bz_debugMessage(debuglevel,errormsg.c_str());
			}
			else
			{
				limits.push_back(newlimit);
			}
		}
	}
	else
	{
		std::string errormsg = "killLimits: warning: invalid parameter \"-" + arg + "\" (ignoring)";
		bz_debugMessage(debuglevel,errormsg.c_str());
	}
	delete params;
}

void killLimits::removeempties(std::vector<std::string>& arr)
{
	for (std::vector<std::string>::iterator x = arr.begin(); x != arr.end(); x++)
	{
		if (*x == "")
		{
			arr.erase(x);
		}
	}
}

void killLimits::trim(std::string& str)
{
	if (str == "")
	{
		return;
	}

	//trim the back end
	int x;
	for (x = str.length() - 1; x >= 0; x--)
	{
		if (str[x] != ' ')
		{
			if (x != str.length() - 1)
			{
				str = str.substr(0, x + 1);
			}
			break;
		}
	}
	if (x < 0)
	{
		str = ""; //this string is all spaces.
		return;
	}
	//and the front...
	for (unsigned int y = 0; y < str.length(); y++)
	{
		if (str[y] != ' ')
		{
			if (y != 0)
			{
				str = str.substr(y);
			}
			break;
		}
	}
}

std::vector<std::string>* killLimits::explode(std::string src, char s)
{
	std::vector<std::string>* list = new std::vector<std::string>;

	int index = 0;
	list->push_back("");

	for (unsigned int x = 0; x < src.length(); x++)
	{
		if (src[x] == s)
		{
			index++;
			list->push_back("");
		}
		else
		{
			list->at(index) += src[x];
		}
	}
	return list;
}

int killLimits::strtoint (std::string src)
{
	int val = 0;
	int digit;
	for (unsigned int x = 0; x < src.length(); x++)
	{
		if (src[x] < '0' || src[x] > '9')
		{
			return 0; //not an integer
		}
		digit = static_cast<int> (src[x] - '0');
		val *= 10;
		val += digit;
	}
	return val;
}

int killLimits::totaltanks()
{
	return 
		bz_getTeamCount (eRogueTeam) +
		bz_getTeamCount (eRedTeam) +
		bz_getTeamCount (eGreenTeam) +
		bz_getTeamCount (eBlueTeam) +
		bz_getTeamCount (ePurpleTeam) +
		bz_getTeamCount (eRabbitTeam);
}

std::string killLimits::lookupflag(int playerID)
{
	bz_PlayerRecord* playerinfo = bz_getPlayerByIndex(playerID);
	if (playerinfo != NULL)
	{
		std::string flag = abbrev(std::string(playerinfo->currentFlag.c_str()));
		bz_freePlayerRecord (playerinfo);
		return flag;
	}
	return std::string("");
}

std::string killLimits::abbrev(std::string src)
{
	if (src.length() < 4)
	{
		return std::string("");
	}
	std::string trimmed = src;

	if (trimmed == "Red team flag")
	{
		return std::string("R*");
	}
	else if (trimmed == "Green team flag")
	{
		return std::string("G*");
	}
	else if (trimmed == "Blue team flag")
	{
		return std::string("B*");
	}
	else if (trimmed == "Purple team flag")
	{
		return std::string("P*");
	}

	//get rid of any whitespace
	trim(trimmed);

	//remove parenthesis at the end
	trimmed = trimmed.substr(0, trimmed.length() - 1);
	
	//snip off everything else
	trimmed = trimmed.substr(trimmed.length() - 2);

	//if the abbreviation is one letter, remove the + or ( or -...
	if (trimmed[0] == '+' || trimmed[0] == '(' || trimmed[0] == '-')
	{
		trimmed = trimmed.substr(1);
	}

	return trimmed;
}

// Local Variables: ***
// mode:C++ ***
// tab-width: 8 ***
// c-basic-offset: 2 ***
// indent-tabs-mode: t ***
// End: ***
// ex: shiftwidth=2 tabstop=8

